import 'package:apex_flutter_plugin/apex_flutter_plugin.dart';
import 'package:apex_flutter_plugin/auth/login.dart';
import 'package:apex_flutter_plugin/auth/user_verification.dart';
import 'package:fingerprintjs/fingerprintjs.dart'
    if (dart.library.io) 'package:apex_flutter_plugin/fingerprint_web.dart';
import 'package:flutter/foundation.dart';
import 'package:flutter/material.dart';
import 'package:shared_preferences/shared_preferences.dart';

import 'api/config.dart';

typedef ActionType = Map<Type, Function(Map<String, dynamic>)>?;

class BaseAction {
  final int _success;
  final String _message;
  final Map<String, dynamic>? rawData;
  final LoginStep step;

  String get message => _message.toString();

  int get success => _success;

  @mustCallSuper
  BaseAction(this._success, this._message, {this.rawData})
      : step = _success == -4
            ? LoginStep.showOtp
            : _success == -5
                ? LoginStep.showPassword
                : _success == -3
                    ? LoginStep.showUsername
                    : _success == -2
                        ? LoginStep.showUpdate
                        : _success == 1
                            ? LoginStep.success
                            : LoginStep.failure;

  factory BaseAction.fromJson(Map<String, dynamic> json) {
    return BaseAction(json['success'],
        json['message'] != null ? json['message'].toString() : '',
        rawData: json);
  }
}

class _Server extends InheritedWidget {
  final ApiConfig config;
  final ActionType _actions;
  final void Function(bool open)? dialogController;

  _Server(
      {Key? key,
      required Widget child,
      required this.config,
      required ActionType actions,
      this.dialogController})
      : _actions = actions!
          ..addAll({
            BaseAction: (x) => BaseAction.fromJson(x),
            LoginAction: (x) => LoginAction.fromJson(x),
            UserVerificationAction: (x) => UserVerificationAction.fromJson(x),
          }),
        super(key: key, child: child) {
    Api.shared.config = config;
  }

  Future<K> makeRequest<K extends BaseAction>(Request request,
      {bool showDialog = false}) async {
    if (dialogController != null && showDialog) dialogController!(true);
    Response? response;
    var error;

    if (request.isPrivate &&
        !request.has('token') &&
        ![1001, 1002, 1003, 1004].contains(request.action)) {
      SharedPreferences prefs = await SharedPreferences.getInstance();
      final token = prefs.get('apex_api_token');
      if (token != null) request.addParam(key: 'token', value: token);
    }

    request.addParam(
        key: 'fingerprint', value: await ApexFlutterPlugin.hashUID());

    try {
      response = await Api.shared.makeRequest(IsolateModel(request));
    } catch (e) {
      error = e;
    } finally {
      if (dialogController != null && showDialog) dialogController!(false);
    }

    if (response != null)
      return _actions![K]!(response.result);
    else
      return Future.error(error);
  }

  // static _Server of(BuildContext context) {
  //   final _Server? result =
  //       context.dependOnInheritedWidgetOfExactType<_Server>();
  //   assert(result != null, 'No Server found in context');
  //   return result!;
  // }

  @override
  bool updateShouldNotify(_Server old) => config != old.config;
}

class ServerWidget extends StatefulWidget {
  static _Server of(BuildContext context) =>
      context.dependOnInheritedWidgetOfExactType<_Server>() as _Server;

  final Widget child;
  final Widget? loading;
  final ApiConfig config;
  final Map<Type, void Function(Map<String, dynamic> json)> actions;

  const ServerWidget(
      {Key? key,
      required this.child,
      this.loading,
      required this.config,
      required this.actions})
      : super(key: key);

  @override
  _ServerWidgetState createState() => _ServerWidgetState();
}

class _ServerWidgetState extends State<ServerWidget> {
  bool showingProgress = false;

  @override
  void initState() {
    super.initState();

    prepareAdditionalParameters();
  }

  @override
  Widget build(BuildContext context) {
    return _Server(
      actions: widget.actions,
      config: widget.config,
      dialogController: (open) {
        showingProgress = open;
        setState(() {});
      },
      child: Material(
        type: MaterialType.transparency,
        child: Directionality(
          textDirection: TextDirection.ltr,
          child: Stack(
            children: [
              IgnorePointer(ignoring: showingProgress, child: widget.child),
              if (showingProgress)
                Container(
                    color: Colors.black26,
                    alignment: Alignment.center,
                    child: widget.loading ??
                        Container(
                          width: 100,
                          height: 100,
                          decoration: BoxDecoration(
                            borderRadius: BorderRadius.circular(16),
                            color: Colors.white,
                          ),
                          child: Center(
                            child: SizedBox(
                              width: 48,
                              height: 48,
                              child: CircularProgressIndicator(),
                            ),
                          ),
                        ))
            ],
          ),
        ),
      ),
    );
  }

  void prepareAdditionalParameters() async {
    // if (defaultTargetPlatform == TargetPlatform.android) {
    // try {
    //   Map<Permission, PermissionStatus> statuses = await [
    //     Permission.phone,
    //   ].request();
    //
    //   if (statuses.values.every((ps) => ps.isGranted)) {
    //     try {
    //       widget.config.defaultParameters?['additional']['imei'] = await ImeiImsiPlugin.getImei(
    //           shouldShowRequestPermissionRationale: false);
    //       widget.config.defaultParameters?['additional']['imsi'] =
    //           await ImeiImsiPlugin.getImsi(
    //               shouldShowRequestPermissionRationale: false);
    //     } on PlatformException {}
    //     widget.config.defaultParameters!["fingerprint"] = await ApexFlutterPlugin.hashUID();
    //   } else {
    //     exit(0);
    //   }
    // } on PlatformException {
    //   widget.config.defaultParameters!["fingerprint"] = sha256
    //       .convert(utf8.encode((await ApexFlutterPlugin.fingerprint())!))
    //       .toString();
    // }
    // }

    if (kIsWeb) {
      String fingerprint = await Fingerprint.getHash();
      widget.config.defaultParameters!["fingerprint"] = fingerprint;
    }
  }
}

enum LoginStep {
  showOtp,
  showUsername,
  showPassword,
  showUpdate,
  success,
  failure
}
